const { expect } = require('chai')
const sinon = require('sinon')
const DependencyInjector = require('../src/DependencyInjector')

describe('the dependency injector', function() {
    it('should be a functional class', function() {
        expect(DependencyInjector).to.be.a('function')
    })

    it('should be constructable', function(){
        expect(new DependencyInjector()).to.be.an.instanceof(DependencyInjector)
    })

    it('should set the container on construction', function() {
        const fake_container = {}
        const di = new DependencyInjector(fake_container)
        expect(di.container).to.be.equal(fake_container)
    })

    it('should inject itself into the container', function() {
        const fake_container = {}
        const di = new DependencyInjector(fake_container)
        expect(fake_container.di).to.not.be.undefined
        expect(fake_container.di).to.be.equal(di)
    })

    describe('inject function', function() {
        let container
        let di
        let fake_class
        beforeEach(function() {
            container = {}
            di = new DependencyInjector(container)
            di.__is_injectable = sinon.fake.returns(true)
            fake_class = { __inject: sinon.spy(), name: 'fake_class' }
        })

        it('should check if the class is injectable', function() {
            di.inject(fake_class)
            expect(di.__is_injectable.callCount).to.be.equal(1)
        })

        it('should fail if the class is not injectable', function() {
            di.__is_injectable = sinon.fake.returns(false)
            expect(di.inject.bind(di, fake_class)).to.throw('Cannot inject non-injectable class: '+fake_class.name)
        })

        it('should call the inject method on class', function() {
            di.inject(fake_class)
            expect(fake_class.__inject.calledOnce).to.be.true
        })

        it('should pass the container to the inject method', function() {
            di.inject(fake_class)
            expect(fake_class.__inject.getCall(0).args.length).to.be.equal(1)
            expect(fake_class.__inject.getCall(0).args[0]).to.be.equal(container)
        })

        it('should return the class', function() {
            expect(di.inject(fake_class)).to.be.equal(fake_class)
        })
    })

    describe('make function', function() {
        let container
        let di
        let fake_class
        beforeEach(function() {
            container = {}
            di = new DependencyInjector(container)
            di.__is_injectable = sinon.fake.returns(true)
            di.inject = sinon.spy()
        })

        it('should be a function', function() {
            expect(di.make).to.be.a('function')
        })

        it('should expect at least one argument', function() {
            expect(di.make.length).to.be.equal(1)
        })

        it('should instantiate classes', function() {
            class Test {}
            expect(di.make(Test)).to.be.an.instanceof(Test)
        })

        it('should pass along constructor arguments', function() {
            class Test {
                constructor(foo, bar) {
                    this.foo = foo
                    this.bar = bar
                }
            }

            const sym_1 = Symbol()
            const sym_2 = Symbol()
            const test = di.make(Test, sym_1, sym_2)
            expect(test.foo).to.be.equal(sym_1)
            expect(test.bar).to.be.equal(sym_2)
        })

        it('should inject injectable classes', function() {
            class Test {}
            di.make(Test)
            expect(di.inject.calledOnce).to.be.true
            expect(di.inject.getCall(0).args.length).to.be.equal(1)
            expect(di.inject.getCall(0).args[0]).to.be.equal(Test)
        })

        it('should not inject normal classes', function() {
            class Test {}
            di.__is_injectable = sinon.fake.returns(false)
            di.make(Test)
            expect(di.inject.calledOnce).to.be.false
        })
    })

    describe('get function', function() {
        let container
        let di
        beforeEach(function() {
            container = { get: sinon.spy() }
            di = new DependencyInjector(container)
        })

        it('should be a function', function() {
            expect(di.get).to.be.a('function')
        })

        it('should expect one argument', function() {
            expect(di.get.length).to.be.equal(1)
        })

        it('should call the service method on the container', function() {
            const arg = {}
            di.get(arg)
            expect(container.get.calledOnce).to.be.true
            expect(container.get.getCall(0).args.length).to.be.equal(1)
            expect(container.get.getCall(0).args[0]).to.be.equal(arg)
        })
    })

    describe('[DEPRECATED] service function', function() {
        let container
        let di
        beforeEach(function() {
            container = { get: sinon.spy() }
            di = new DependencyInjector(container)
        })

        it('should be a function', function() {
            expect(di.service).to.be.a('function')
        })

        it('should expect one argument', function() {
            expect(di.service.length).to.be.equal(1)
        })

        it('should call the service method on the container', function() {
            const arg = {}
            di.service(arg)
            expect(container.get.calledOnce).to.be.true
            expect(container.get.getCall(0).args.length).to.be.equal(1)
            expect(container.get.getCall(0).args[0]).to.be.equal(arg)
        })
    })

    describe('has function', function() {
        let container
        let di
        beforeEach(function() {
            container = { has: sinon.spy() }
            di = new DependencyInjector(container)
        })

        it('should be a function', function() {
            expect(di.has).to.be.a('function')
        })

        it('should expect 1 argument', function() {
            expect(di.has.length).to.be.equal(1)
        })

        it('should call the has method on the container', function() {
            const name = Math.random().toString(36).substring(7)
            di.has(name)
            expect(container.has.calledOnce).to.be.true
            expect(container.has.getCall(0).args.length).to.be.equal(1)
            expect(container.has.getCall(0).args[0]).to.be.equal(name)
        })

        it('should call the has method on the container for each element in the argument array, one is passed', function() {
            const n_items = parseInt(String(Math.random() * 100))
            const arr = Array(n_items).fill('')
            container.has = sinon.fake.returns(true)
            expect(di.has(arr)).to.be.true
            expect(container.has.getCalls().length).to.be.equal(n_items)
        })
    })

    describe('inject_globally function', function() {
        it('should be a function', function() {
            const di = new DependencyInjector()
            expect(di.inject_globally).to.be.a('function')
        })

        it('should accept no arguments', function() {
            const di = new DependencyInjector()
            expect(di.inject_globally.length).to.be.equal(0)
        })

        it('should replace the global require method', function() {
            const mod_proto = require('module').prototype
            const original_require = mod_proto.require
            const di = new DependencyInjector()
            di.inject_globally()
            expect(mod_proto.require).to.not.be.equal(original_require)
            mod_proto.require = original_require
        })

        it('should call the inject function for injectable classes when they are loaded through require', function() {
            const mod = require('module')
            const original_require = mod.prototype.require

            const sym_1 = Symbol()
            mod.prototype.require = what => {
                if ( what === 'module' ) return mod
                else return sym_1
            }

            const di = new DependencyInjector()
            const sym_2 = Symbol()
            di.inject = sinon.fake.returns(sym_2)
            di.__is_injectable = sinon.fake.returns(true)

            di.inject_globally()

            const test_val = require('this should not do anything')
            expect(di.inject.calledOnce).to.be.true
            expect(di.inject.getCall(0).args[0]).to.be.equal(sym_1)
            expect(test_val).to.be.equal(sym_2)

            mod.prototype.require = original_require
        })

        it('should not call the inject function if requires are not injectable', function() {
            const mod = require('module')
            const original_require = mod.prototype.require

            const sym_1 = Symbol()
            mod.prototype.require = what => {
                if ( what === 'module' ) return mod
                else return sym_1
            }

            const di = new DependencyInjector()
            const sym_2 = Symbol()
            di.inject = sinon.fake.returns(sym_2)
            di.__is_injectable = sinon.fake.returns(false)

            di.inject_globally()

            const test_val = require('this should not do anything')
            expect(di.inject.calledOnce).to.be.false
            expect(test_val).to.be.equal(sym_1)

            mod.prototype.require = original_require
        })
    })

    describe('__is_injectable function', function() {
        let di
        let SomeClass
        beforeEach(function() {
            di = new DependencyInjector()
            class TestClass {
                static __inject(a) {}
            }
            SomeClass = TestClass
        })

        it('should be a function', function() {
            expect(di.__is_injectable).to.be.a('function')
        })

        it('should accept one argument', function() {
            expect(di.__is_injectable.length).to.be.equal(1)
        })

        it('should fail if the class is not functional', function() {
            expect(di.__is_injectable(SomeClass)).to.be.true
            SomeClass = {}
            expect(di.__is_injectable(SomeClass)).to.be.false
        })

        it('should fail if the class does not have a static __inject function', function() {
            expect(di.__is_injectable(SomeClass)).to.be.true
            SomeClass.__inject = true
            expect(di.__is_injectable(SomeClass)).to.be.false
        })

        it('should fail if __inject does not take at least one argument', function() {
            expect(di.__is_injectable(SomeClass)).to.be.true
            SomeClass.__inject = function() {}
            expect(di.__is_injectable(SomeClass)).to.be.false
        })
    })
})
